package animation;

import java.util.Stack;

import animation.StackFrame.FrameType;

public class Memory {

    public enum MemoryType
    {
        GLOBAL,
        LOCAL,
        CURRENT
    }
    
    private StackFrame current;
    private StackFrame global;
    private Stack< StackFrame > stack;
    
    public Memory()
    {
        stack = new Stack<StackFrame>();
        global = new StackFrame( FrameType.FUNCTION );
        current = null;
    }
    
    public int getSize()
    {
        return stack.size();
    }
    
    public void setCurrent( StackFrame frame )
    {
        this.current = frame;
    }
    
    public void addFrame( StackFrame frame )
    {
        stack.push( frame );
    }
    
    public StackFrame removeFrame()
    {
        return stack.pop();
    }
    
    public <T> void declare( String name, T value, MemoryType type )
    {
        switch( type )
        {
        case GLOBAL:
            this.global.declare( name, value );
            break;
        case CURRENT:
            if( current != null )
                current.declare( name, value );
            break;
        case LOCAL:
        	if( stack.size() == 0 )
        		return;
            stack.peek().declare( name, value );
            break;
        }
    }
    
    public <T> void write( String name, T value, MemoryType type )
    {
        switch( type )
        {
        case GLOBAL:
            this.global.set( name, value );
            break;
        case CURRENT:
            if( current != null )
            {
                int index = stack.size() - 1;
                while( index >= 0 && stack.get( index ) != current ) {
                    index--;
                }
                if( index < 0 )
                    break;
                while( index >= 0 ) {
                    StackFrame stackF = stack.get( index-- );
                    if( stackF.isDefined( name ) )
                    {
                        stackF.set( name, value );
                        return;
                    }
                    if( stackF.getType() == FrameType.FUNCTION )
                        break;
                 }
            }
            break;
        case LOCAL:
        	int index = stack.size() - 1;
        	while( index >= 0 ) {
        	   StackFrame stackF = stack.get( index-- );
        	   if( stackF.isDefined( name ) )
        	   {
        		   stackF.set( name, value );
        		   return;
        	   }
        	   if( stackF.getType() == FrameType.FUNCTION )
        		   break;
        	}
        }
    }
    
    public <T> T read( String name, MemoryType type )
    {
        switch( type )
        {
        case GLOBAL:
            return this.global.get( name );
        case CURRENT:
            if( current != null )
            {
                int index = stack.size() - 1;
                while( index >= 0 && stack.get( index ) != current ) {
                    index--;
                }
                if( index < 0 )
                    break;
                while( index >= 0 ) {
                    StackFrame stackF = stack.get( index-- );
                    if( stackF.isDefined( name ) )
                        return stackF.get( name );
                    if( stackF.getType() == FrameType.FUNCTION )
                        break;
                 }
            }
            break;
        case LOCAL:
            int index = stack.size() - 1;
            while( index >= 0 ) {
                StackFrame stackF = stack.get( index-- );
                if( stackF.isDefined( name ) )
                    return stackF.get( name );
                if( stackF.getType() == FrameType.FUNCTION )
                    break;
             }
        }
        return null;
    }
    
    public boolean isSomewhereDefined( String name, MemoryType type )
    {
        switch( type )
        {
        case GLOBAL:
            return this.global.isDefined( name );
        case CURRENT:
            if( current != null )
            {
                int index = stack.size() - 1;
                while( index >= 0 && stack.get( index ) != current ) {
                    index--;
                }
                if( index < 0 )
                    break;
                while( index >= 0 ) {
                    StackFrame stackF = stack.get( index-- );
                    if( stackF.isDefined( name ) )
                        return true;
                    if( stackF.getType() == FrameType.FUNCTION )
                        break;
                 }
            }
            break;
        case LOCAL:
            int index = stack.size() - 1;
            while( index >= 0 ) {
                StackFrame stackF = stack.get( index-- );
                if( stackF.isDefined( name ) )
                    return true;
                if( stackF.getType() == FrameType.FUNCTION )
                    break;
             }
        }
    	return false;
    }
    
    public boolean isDefined( String name, MemoryType type )
    {
        switch( type )
        {
        case GLOBAL:
            return this.global.isDefined( name );
        case CURRENT:
            if( current != null)
                return current.isDefined( name );
            return false;
        case LOCAL:
            if( stack.size() == 0 )
                return false;
            return stack.peek().isDefined( name );
        }
    	return false;
    }
    
    public void undeclare( String name, MemoryType type )
    {
        switch( type )
        {
        case GLOBAL:
            this.global.undeclare( name );
            break;
        case CURRENT:
            if( current != null )
                current.undeclare( name );
            break;
        case LOCAL:
        	if( stack.size() == 0 )
        		return;
            stack.peek().undeclare( name );
        }
    }
}
